class Selection:
    """
    Selection of trees for crossover and mutation

    Args:
        population (list): List of trees.
        selection_tournament_size (int): Number of trees to select.
        random (Random): Random number generator.
    """
    def selection(population, selection_tournament_size, random):
        """
        Select two trees for crossover or mutation.

        Args:
            population (list): List of trees.
            selection_tournament_size (int): Number of trees to select.
            random (Random): Random number generator.

        Returns:
            tuple: The two selected trees.
        """
        valid = False
        while not valid:
            selection = []

            # Select two trees
            for _ in range(2):
                indices = random.choice(len(population),
                                        selection_tournament_size, replace=False)
                candidates = [population[i] for i in indices]
                candidates.sort(key=lambda x: x.fitness, reverse=True)
                selection.append(candidates[0])

            # Check if trees are different
            if selection[0] != selection[1]:
                valid = True

        return selection[0], selection[1]



class SettingOne:
    def __init__(self):
        self.RUN_COUNT = 10  # Number of individual runs
        self.POPULATION_SIZE = 150  # Number of individuals in one generation
        self.ELITE_SIZE = 1  # Elite automatically advances to the next generation. If -1, then children compete with parents, and only the best advance
        self.GENERATION_LIMIT = 2000  # Number of generations
        # Stop evolution when the best individual stays the same for x generations
        self.STAGNATION_LIMIT = 50

        # -------- MUTATION -------------------------
        # 0 = STANDARD
        # 1 = SMART - WORST NODE
        self.MUTATION_METHOD = 0
        self.MUTATION_PROBABILITY = 0.1
        # self.MUTATION_ATTRIBUTE_PROB = 0.03  # mutation probability for attribute nodes
        # self.MUTATION_SPLIT_PROB = 0.07  # mutation probability for split border
        # self.MUTATION_DECISION_PROB = 0.09  # mutation probability for decision nodes
        # self.MUTATION_ATTRIBUTE_DECISION_PROB = 0.01  # mutation probability for exchange of attribute to decision nodes
        # self.MUTATION_DECISION_ATTRIBUTE_PROB = 0.01  # mutation probability for exchange of decision to attribute nodes

        # -------- SCORES --------------------------
        self.PENALTY_NUM_NODES = 0.002  # penalty score for each node
        # penalty score for each unused decision node
        self.PENALTY_NUM_UNUSED_NODES = 0.02
        self.PENALTY_DYNAMIC = 20  # penalty factor for dynamic fitness

        # -------- SELECTION -----------------------
        # 0 = ROULETTE
        # 1 = LINEAR RANKING
        # 2 = LOG RANKING
        # 3 = TOURNAMENT
        self.SELECTION_METHOD = 3
        self.SELECTION_TOURNAMENT_SIZE = 2

        # -1= NO SMART SELECTION
        # 0 = SIZE COMPARISON
        # 1 = ATTRIBUTE COMPARISON
        # 2 = ATTRIBUTE VALUE COMPARISON
        # 3 = CONFUSION MATRIX COMPARISON
        # 4 = CLASSIFICATION DIF. COMPARISON
        # 5 = FAMILY SELECTION (prevent inbreeding; use with FAMILY TREE LEVELS)
        self.SELECTION_SMART_METHOD = -1

        # -------- CALCULATOR ----------------------
        # 0 = ACCURACY
        # 1 = VILI (accuracy with size and unused penalty)
        # 2 = FSCORE (average fscore with size and unused penalty)
        # 3 = PERCLASS (accuracy per class with size and unused penalty)
        # 4 = PERCLASSINVERSE (accuracy with size and unused penalty)
        # (smaller the percentage of class more weight it has)
        # 5 = AVERAGE CLASS ACCURACY
        # 6 = DYNAMIC FSCORE : fit = (1-fsc) + numNodes / (numInstances * PENALTY_DYNAMIC)
        # 7 = DYNAMIC AVG CLASS ACCURACY : fit = (1-fsc) + numNodes / (numInstances * PENALTY_DYNAMIC)
        # 8 = SAME AS 7, BUT CLASS ACCURACIES ARE WEIGHTED
        self.CALCULATOR_METHOD = 6

        # -------- CROSSOVER -----------------------
        # 0 = STANDARD
        # 1 = VILI
        # 2 = ACCURACY
        # 3 = FSCORE ALL (ZMNOZEK VSEH)
        # 4 = FSCORE - REPLACE WORST FSCORE WITH BEST
        # 5 = ONLY NODE - AS STANDARD BUT SUBTREES ARE NOT EXCHANGED
        # 6 = FSCORE AVERAGE - REPLACES THE WORST AVERAGE FSCORED NODE WITH THE BEST ONE
        # 7 = FSCORE WEIGHTED - REPLACES THE WORST WEIGHTED FSCORED NODE WITH THE BEST ONE
        # 8 = ACCURACY AND USAGE OF NODES (0.8 acc + 0.4 (percentage of instances that go through that node))
        # 9 = ACCURACY AND USAGE OF NODES (0.8 acc * 0.4 (percentage of instances that go through that node))
        # 10= SAME AS 8, BUT REPLACES WORST WITH RANDOM (BEST SO FAR)
        # 11= SAME AS 9, BUT REPLACES WORST WITH RANDOM
        # 12= NODE USAGE AVG
        # 13= NODE USAGE SUM
        # -1= NO CROSSOVER
        self.CROSSOVER_METHOD = 0

        # If(True) crossovers 2,3,4 don't settle with worst leaves, but they exchange subtree with those leaves
        self.CROSSOVER_ONLYLEAVES = False

        # Maximum depth of the trees in randomly generated trees in the starting population. -1 makes MAX_DEPTH J48 result * 2 depth.
        self.MAX_DEPTH = -1

        # -------- DUPLICATES -------------------
        # 0 = allow duplicates
        # 1 = remove duplicates after mutation (population is shrunk for 1 generation)
        # 2 = remove duplicates, but maintain the population size all the time
        self.DUPLICATE_REMOVER = 0

        # -------- FAMILY TREE -----------------
        self.FAMILY_TREE_LEVELS = 0  # Levels for family trees (0-off)

        # -------- MULTI POPULATIONS ----------
        self.MULTI_POP_NUM = 1  # Number of independent populations (islands)
        self.MULTI_MIGRATION_STEP = 100  # Every X generation migration happens
        self.MULTI_MIGRATION_SIZE = 10  # How many emigrate to another population

        # ------- ENSEMBLES -------------------
        # -1 = NO BAGGING
        # 0  = BAGGING, RANDOM
        # 1  = BAGGING, ACCORDING TO WEIGHTS (USE WITH BOOSTING)
        # 2  = RANDOM FOREST
        self.BAGGING_METHOD = -1
        self.BAGGING_STEP = 200

        # -1 = NO BOOSTING
        # 0  = BOOSTING, RESETS EVOLUTION AFTER EVERY BOOST
        # 1  = BOOSTING, CONTINUE EVOLUTION AFTER EVERY BOOST
        # 2  = SAME AS 1, BUT NO DUPLICATED CLASSIFIERS IN ENSEMBLE
        # 3  = LOGITBOOST
        self.BOOSTING_METHOD = -1
        self.BOOSTING_STEP = 200

        # ------- FITNESS SKIPPING -------------------
        # Fitness is not calculated X number of generations.
        # Selection is made based on the family trees.
        # Use in combination with FAMILY TREE LEVELS.
        # -1 = Off (don't use).
        # Fitness is not calculated X number of generations
        self.FITNESS_SKIP_GENERATIONS = -1

        # ------- PROBLEM TYPE -----------------------
        # 0  = Classification
        # 1  = Clustering
        # 2  = Regression
        self.PROBLEM_TYPE = 0

        # ------- VALIDATION SET -----------------------
        self.VALIDATION_SET_PERCENTAGE = 0.5
        self.VALIDATION_SET_WEIGHT = 1

    def __str__(self):
        result = "{"
        for attr, value in self.__dict__.items():
            result += f"{attr}: {value}, "
        result = result[:-2] + "}"
        return result

    def main(self):
        s = SettingOne()
        print(s)



import numpy as np
from joblib import Parallel, delayed
from sklearn.metrics import accuracy_score

# GATree
from gatree.tree.node import Node
from gatree.ga.selection import Selection
from gatree.ga.crossover import Crossover
from gatree.ga.mutation import Mutation


class GATree():
    """
    Genetic Algorithm Tree Classifier.

    Args:
        max_depth (int, optional): Maximum depth of the tree.
        random (Random, optional): Random number generator.
        fitness_function (function, optional): Fitness function for the genetic algorithm.
        n_jobs (int, optional): Number of jobs to run in parallel.
        random_state (int, optional): Seed for reproducibility.
        _tree (Node, optional): The fitted tree.
        _best_fitness (list, optional): List of best fitness values for each iteration.
        _avg_fitness (list, optional): List of average fitness values for each iteration.

    Attributes:
        max_depth (int, optional): Maximum depth of the tree.
        random (Random): Random number generator.
        X (pandas.DataFrame): Training data.
        y (pandas.Series): Target values.
        att_indexes (numpy.ndarray): Array of attribute indexes.
        att_values (dict): Dictionary of attribute values.
        class_count (int): Number of classes.
        fitness_function (function): Fitness function for the genetic algorithm.
        n_jobs (int): Number of jobs to run in parallel.
        random_state (int): Seed for reproducibility.
        _tree (Node): The fitted tree.
        _best_fitness (list): List of best fitness values for each iteration.
        _avg_fitness (list): List of average fitness values for each iteration.
    """

    def __init__(self, max_depth=None, random=None, fitness_function=None, n_jobs=1, random_state=None):
        """
        Initialize the Genetic Algorithm Tree Classifier.

        Args:
            max_depth (int, optional): Maximum depth of the tree.
            random (Random, optional): Random number generator.
            fitness_function (function, optional): Fitness function for the genetic algorithm.
            random_state (int, optional): Seed reproducibility.
        """
        self.max_depth = max_depth
        if random is None and random_state is not None:
            np.random.seed(random_state)
        self.random = random if random is not None else np.random
        self.fitness_function = fitness_function if fitness_function is not None else self.default_fitness_function
        self.n_jobs = n_jobs
        self._tree = None
        self._best_fitness = []
        self._avg_fitness = []

    @staticmethod
    def default_fitness_function(root):
        """ 
        Default fitness function for the genetic algorithm.

        Args:
            root (Node): Root node of the tree.

        Returns:
            float: The fitness value.
        """
        return accuracy_score(root.y_true, root.y_pred) - 0.002 * root.size()

    @staticmethod
    def _predict_and_evaluate(tree, X, y, fitness_function):
        """
        Evaluate a tree on a training set (in parallel)

        Args:
            tree (Node): Tree to evaluate.
            X (pandas.DataFrame): Training data.
            y (pandas.Series): Target values.
            fitness_function (function): Fitness function for the genetic algorithm.

        Returns:
            Node: The evaluated tree.
        """
        for j in range(X.shape[0]):
            # Predict class for current instance
            tree.predict_one(X.iloc[j], y.iloc[j])
        tree.fitness = fitness_function(tree)
        return tree

    def fit(self, X, y, population_size=150, max_iter=2000, mutation_probability=0.1, elite_size=1, selection_tournament_size=2):
        """
        Fit a tree to a training set.

        Args:
            X (pandas.DataFrame): Training data.
            y (pandas.Series): Target values.
            population_size (int, optional): Size of the population.
            max_iter (int, optional): Maximum number of iterations.
            mutation_probability (float, optional): Probability of mutation.
            elite_size (int, optional): Number of elite trees.
            selection_tournament_size (int, optional): Number of trees in tournament.

        Returns:
            Node: The fitted tree.
        """
        self.X = X
        self.y = y
        self.att_indexes = np.arange(X.shape[1])
        self.att_values = {i: [(min_val + max_val) / 2 for min_val, max_val in zip(sorted(
            X.iloc[:, i].unique())[:-1], sorted(X.iloc[:, i].unique())[1:])] for i in range(X.shape[1])}
        self.att_values[-1] = sorted(y.unique())
        self.class_count = len(self.att_values[-1])

        # Generation of initial population
        node = Node()
        population = []
        for _ in range(population_size):
            population.append(node.make_node(max_depth=self.max_depth, random=self.random,
                              att_indexes=self.att_indexes, att_values=self.att_values, class_count=self.class_count))

        for i in range(max_iter+1):
            # Clear previous evaluation
            for tree in population:
                tree.clear_evaluation()

            # Evaluation of population
            population = Parallel(n_jobs=self.n_jobs)(delayed(GATree._predict_and_evaluate)(
                tree, X, y, self.fitness_function) for tree in population)

            # Sort population by fitness
            population.sort(key=lambda x: x.fitness, reverse=True)

            # Log best and average fitness
            self._best_fitness.append(population[0].fitness)
            self._avg_fitness.append(
                sum([tree.fitness for tree in population]) / len(population))

            if i != max_iter:
                # Elites
                elites = population[:elite_size]

                # Descendant generation
                descendant = []
                for _ in range(0, len(population), 2):
                    # Tournament selection
                    tree1, tree2 = Selection.selection(
                        population=population, selection_tournament_size=selection_tournament_size, random=self.random)

                    # Crossover between selected trees
                    crossover1 = Crossover.crossover(
                        tree1=tree1, tree2=tree2, random=self.random)
                    crossover2 = Crossover.crossover(
                        tree1=tree2, tree2=tree1, random=self.random)

                    # Mutation of new trees
                    mutation1 = crossover1
                    mutation2 = crossover2
                    if self.random.random() < mutation_probability:
                        mutation1 = Mutation.mutation(root=crossover1, att_indexes=self.att_indexes,
                                                      att_values=self.att_values, class_count=self.class_count, random=self.random)
                    if self.random.random() < mutation_probability:
                        mutation2 = Mutation.mutation(root=crossover2, att_indexes=self.att_indexes,
                                                      att_values=self.att_values, class_count=self.class_count, random=self.random)

                    # Add new trees to descendant population
                    descendant.extend([mutation1, mutation2])

                # Elites + descendants
                descendant.sort(key=lambda x: x.fitness, reverse=True)
                descendant = elites + descendant[:population_size - elite_size]

                # Replace old population with new population
                population = descendant

        self._tree = population[0]

    def predict(self, X):
        """
        Predict classes for the given data.

        Args:
            X (pandas.DataFrame): Data to predict.

        Returns:
            list: Predicted classes.
        """
        y_pred = []
        for i in range(X.shape[0]):
            index = self._tree.predict_one(X.iloc[i])
            y_pred.append(self.att_values[-1][index])
        return y_pred

    def plot(self, node=None, prefix=''):
        """
        Plot the decision tree with nodes and leaves.

        Args:
            node (Node, optional): Current node to plot.
            prefix (str, optional): Prefix for the current node.
        """
        if node is None:
            node = self._tree

        if node is not None:
            if node.att_index != -1:
                print(prefix + '├── {} > {}'.format(self.X.columns.tolist()
                      [node.att_index], node.att_value))
            else:
                print(
                    prefix + '└── Class: {}'.format(self.att_values[-1][node.att_value]))

            if node.left is not None or node.right is not None:
                self.plot(node.left, prefix + '    ')
                self.plot(node.right, prefix + '    ')



from gatree.tree.node import Node


class Crossover:
    """
    Crossover operations for genetic algorithm.
    """
    def crossover(tree1, tree2, random):
        """
        Perform crossover between two trees.

        Args:
            tree1 (Node): The first tree for crossover.
            tree2 (Node): The second tree for crossover.
            random (Random): Random number generator.

        Returns:
            Node: The new tree resulting from crossover.
        """
        n1 = Node.copy(tree1)
        n2 = Node.copy(tree2)
        size1 = n1.max_depth()
        size2 = n2.max_depth()

        while True:
            if (n1.left == None or random.randint(0, size1) == 0) and n1.parent != None:
                break

            if random.choice([True, False]):
                n1 = n1.left
            else:
                n1 = n1.right

        while True:
            if (n2.left == None or random.randint(0, size2) == 0) and n2.parent != None:
                break

            if random.choice([True, False]):
                n2 = n2.left
            else:
                n2 = n2.right

        p = n1.parent
        if p.left == n1:
            p.set_left(n2)
        else:
            p.set_right(n2)

        return p.get_root()



class Node:
    """
    Node class for decision trees.

    Args:
        att_index (int, optional): Attribute index or -1 if this is a leaf.
        att_value (any, optional): Attribute value or class index if this is a leaf.

    Attributes:
        parent (Node): Parent node.
        left (Node): Left child node.
        right (Node): Right child node.
        att_index (int, optional): Attribute index or -1 if this is a leaf.
        att_value (any, optional): Attribute value or class index if this is a leaf.
        fitness (None): Placeholder for fitness value.
        y_true (list): List of true class values for evaluation.
        y_pred (list): List of predicted class values for evaluation.
    """

    def __init__(self, att_index=None, att_value=None):
        """
        Initializes the node with the given parameters.

        Args:
            att_index (int, optional): Attribute index or -1 if this is a leaf.
            att_value (any, optional): Attribute value or class index if this is a leaf.
        """
        if att_index is None and att_value is not None:
            self.__init__(-1, att_value)  # leaf node
        else:
            self.parent = None
            self.left = None
            self.right = None

            self.att_index = att_index if att_index is not None else None  # if -1 then this is leaf
            self.att_value = att_value if att_value is not None else None

            self.fitness = None
            self.y_true = []
            self.y_pred = []

    @staticmethod
    def copy(node, parent=None):
        """
        Returns a deep copy of the given node.

        Args:
            node (Node): Node to be copied.
            parent (Node, optional): Parent node of the copied node.

        Returns:
            Node: Deep copy of the given node.
        """
        copy = Node(node.att_index, node.att_value)
        copy.parent = parent
        copy.left = Node.copy(
            node.left, parent=copy) if node.left is not None else None
        copy.right = Node.copy(
            node.right, parent=copy) if node.right is not None else None
        copy.fitness = node.fitness
        copy.y_true = node.y_true
        copy.y_pred = node.y_pred
        return copy

    def set_left(self, n):
        """
        Sets the left child of this node to the given node.

        Args:
            n (Node): Node to be set as the left child.
        """
        self.left = n
        n.parent = self

    def set_right(self, n):
        """
        Sets the right child of this node to the given node.

        Args:
            n (Node): Node to be set as the right child.
        """
        self.right = n
        n.parent = self

    def get_root(self):
        """
        Returns the root node of the tree.

        Returns:
            Node: Root node of the tree.
        """
        if self.parent is None:
            return self
        return self.parent.get_root()

    def get_children(self):
        """
        Returns the children of this node.

        Returns:
            tuple: Tuple containing the left and right children nodes.
        """
        return self.left, self.right

    def get_leaves(self):
        """
        Returns the leaves of this node.

        Returns:
            list: List of leaves of this node.
        """
        if self.left is None:
            return [self]
        return self.left.get_leaves() + self.right.get_leaves()

    def depth(self):
        """
        Returns the depth of this node.

        Returns:
            int: Depth of this node.
        """
        if self.parent is None:
            return 1
        return self.parent.depth() + 1

    def max_depth(self):
        """
        Returns the maximum depth of the tree.

        Returns:
            int: Maximum depth of the tree.
        """
        return self.max_depth_helper(self.get_root())

    @staticmethod
    def max_depth_helper(n):
        """
        Helper function for max_depth.

        Args:
            n (Node): Node to be used as the root.

        Returns:
            int: Maximum depth of the tree.
        """
        if n is None:
            return 0
        l_depth = Node.max_depth_helper(n.left)
        r_depth = Node.max_depth_helper(n.right)
        return max(l_depth, r_depth) + 1

    def size(self):
        """
        Size is the number of all nodes (mid-tree nodes + leaves) in the trees.

        Returns:
            int: Number of all nodes in the trees.
        """
        return self.size_helper(self.get_root())

    @staticmethod
    def size_helper(n):
        """
        Helper function for size.

        Args:
            n (Node): Node to be used as the root.

        Returns:
            int: Number of all nodes in the trees.
        """
        if n is None:
            return 0
        return Node.size_helper(n.left) + Node.size_helper(n.right) + 1

    def make_node(self, depth=0, max_depth=None, random=None, att_indexes=None, att_values=None, class_count=None):
        """
        Randomly generates the node and its children.

        Args:
            depth (int, optional): Current depth of the tree.
            max_depth (int, optional): Maximum depth of the tree.
            random (Random, optional): Random number generator.
            att_indexes (numpy.ndarray, optional): Attribute indexes.
            att_values (dict, optional): Attribute values.
            class_count (int, optional): Number of classes.

        Returns:
            Node: Randomly generated node with children.
        """
        node = None
        att_index = None
        value_index = None
        att_value = None
        if max_depth == None:
            max_depth = depth

        try:
            # if it's the root, first level or 50/50 chance of building new children.
            # Must be below maximal depth.
            if (depth <= 1 or (random.choice([True, False])) and depth < max_depth):
                subset_index = random.randint(0, len(att_indexes))
                att_index = att_indexes[subset_index]
                value_index = random.randint(0, len(att_values[att_index]))
                att_value = att_values[att_index][value_index]
                node = Node(att_index=att_index, att_value=att_value)
                node.left = self.make_node(depth=depth + 1, max_depth=max_depth, random=random,
                                           att_indexes=att_indexes, att_values=att_values, class_count=class_count)
                node.left.parent = node
                node.right = self.make_node(depth=depth + 1, max_depth=max_depth, random=random,
                                            att_indexes=att_indexes, att_values=att_values, class_count=class_count)
                node.right.parent = node
            else:  # result (leaf)
                r = random.randint(0, class_count)
                node = Node(att_index=-1, att_value=r)
        except Exception as e:
            print(f"{att_index};{att_value};{value_index}")
            print("Error:", e)
            node = None

        return node

    def clear_evaluation(self):
        """
        Clears the evaluation of this node and all its children.
        """
        self.y_true = []
        self.y_pred = []
        if self.left:
            self.left.clear_evaluation()
        if self.right:
            self.right.clear_evaluation()

    def is_evaluated(self):
        """
        Returns true if this node and all its children are evaluated.

        Returns:
            bool: True if this node and all its children are evaluated.
        """
        if len(self.y_pred) == 0 and len(self.y_true) == 0:
            return False

        left_evaluated = self.left.is_evaluated() if self.left else True
        right_evaluated = self.right.is_evaluated() if self.right else True

        return left_evaluated or right_evaluated

    def predict_one(self, X, y=None):
        """
        Predicts the class of the given instance.

        Args:
            X (list): Instance to be predicted.
            y (int): Actual class of the given instance.

        Returns:
            int: Predicted class.
        """
        try:
            if self.att_index != -1:
                if X.iloc[self.att_index] > self.att_value:
                    if self.left is not None:
                        predicted = self.left.predict_one(X, y)
                else:
                    if self.right is not None:
                        predicted = self.right.predict_one(X, y)
            else:
                predicted = int(self.att_value)

            if y is not None:
                self.y_true.append(int(y))
                self.y_pred.append(predicted)

            return predicted
        except Exception as e:
            print(e)
            return -1

    def __str__(self):
        """
        Returns a string representation of this node.

        Returns:
            str: String representation of this node.
        """
        if self.att_index == -1:
            # Assuming att_value is a numeric index
            return str(int(self.att_value))
            # For nominal values, you might need a mapping from index to value
        else:
            return f"Attribute {self.att_index} > {self.att_value}"

    def __repr__(self):
        """
        Returns a string representation of this node.

        Returns:
            str: String representation of this node.
        """
        return self.__str__()
